#!/usr/bin/env python3
# vim: fileencoding=utf8
# SPDXVersion: SPDX-2.3
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Copyright 2012-2022, 2023 by Christian Dönges
# SPDXID: SPDXRef-unicodechars-py
"""Define characters of the unicode character set by name.

    .. note:: The future of this module is currently unclear. It duplicates
        much of what is in :mod:`unicodedata` from the Python standard
        library. It is a huge effort to make and maintain seeing how large
        ever-growing Unicode already is.

        It may either get spun out into a separate package or made an invisible
        internal part of *pymisclib*.

        You have been warned.

    .. py:data:: UNICODE_CHARS
        :type: dict[str, str]

        A dictionary containing the name of the unicode character as the key
        and the character itself as a string.

        .. deprecated:: 1.5.0 use :mod:`unicodedata` from the Python standard
            library instead.

    .. py:data:: UC
        :type: argparse.Namespace

        A namespace containing namespaces mapped to Unicode blocks.
        The Unicode blocks contain Unicode symbols by name, the value
        is the symbol itself.

        The following blocks are currently defined:

        * arrows
        * block_elements
        * box_drawings

            * double
            * light
            * heavy

        * c0_controls
        * general_punctuation

"""
import unicodedata
from argparse import Namespace

# Dictionary of select Unicode characters, keyed by name.
UNICODE_CHARACTERS = {
    # C0 Code
    "NUL": "\x00",
    "SOH": "\x01",
    "STX": "\x02",
    "ETX": "\x03",
    "EOT": "\x04",
    "ENQ": "\x05",
    "ACK": "\x06",
    "BEL": "\x07",
    "BS": "\x08",  # Backspace
    "HT": "\x09",
    "LF": "\x0a",
    "VT": "\x0b",
    "FF": "\x0c",
    "CR": "\x0d",
    "SO": "\x0e",
    "SI": "\x0f",
    "DLE": "\x10",
    "DC1": "\x11",
    "DC2": "\x12",
    "DC3": "\x13",
    "DC4": "\x14",
    "NAK": "\x15",
    "SYN": "\x16",
    "ETB": "\x17",
    "CAN": "\x18",
    "EM": "\x19",
    "SUB": "\x1a",
    "ESC": "\x1b",
    "FS": "\x1c",
    "GS": "\x1d",
    "RS": "\x1e",
    "US": "\x1f",

    # Control Character
    "DEL": "\x7f",

    "EM SPACE": " ",  # U+2003
    # Arrows, 0x2190-21FF
    "LEFTWARDS ARROW": "←",  # U+2190
    "UPWARDS ARROW": "↑",  # U+2191
    "RIGHTWARDS ARROW": "→",  # U+2192
    "DOWNWARDS ARROW": "↓",  # U+2193
    # Unicode Box Drawing, 0x2500-0x257F
    "BOX DRAWINGS LIGHT HORIZONTAL": "─",  # U+2500
    "BOX DRAWINGS HEAVY HORIZONTAL": "━",  # U+2501
    "BOX DRAWINGS LIGHT VERTICAL": "│",  # U+2502
    "BOX DRAWINGS HEAVY VERTICAL": "┃",  # U+2503
    "BOX DRAWINGS LIGHT TRIPLE DASH HORIZONTAL": "┄",  # U+2504
    "BOX DRAWINGS HEAVY TRIPLE DASH HORIZONTAL": "┅",  # U+2505
    "BOX DRAWINGS LIGHT TRIPLE DASH VERTICAL": "┆",  # U+2506
    "BOX DRAWINGS HEAVY TRIPLE DASH VERTICAL": "┇",  # U+2507
    "BOX DRAWINGS LIGHT QUADRUPLE DASH HORIZONTAL": "┈",  # U+2508
    "BOX DRAWINGS HEAVY QUADRUPLE DASH HORIZONTAL": "┉",  # U+2509
    "BOX DRAWINGS LIGHT QUADRUPLE DASH VERTICAL": "┊",  # U+250A
    "BOX DRAWINGS HEAVY QUADRUPLE DASH VERTICAL": "┋",  # U+250B
    "BOX DRAWINGS LIGHT DOWN AND RIGHT": "┌",  # U+250C
    "BOX DRAWINGS DOWN LIGHT AND RIGHT HEAVY": "┍",  # U+250D
    "BOX DRAWINGS DOWN HEAVY AND RIGHT LIGHT": "┎",  # U+250E
    "BOX DRAWINGS HEAVY DOWN AND RIGHT": "┏",  # U+250F
    "BOX DRAWINGS LIGHT DOWN AND LEFT": "┐",  # U+2510
    "BOX DRAWINGS DOWN LIGHT AND LEFT HEAVY": "┑",  # U+2511
    "BOX DRAWINGS DOWN HEAVY AND LEFT LIGHT": "┒",  # U+2512
    "BOX DRAWINGS HEAVY DOWN AND LEFT": "┓",  # U+2513
    "BOX DRAWINGS LIGHT UP AND RIGHT": "└",  # U+2514
    "BOX DRAWINGS UP LIGHT AND RIGHT HEAVY": "┕",  # U+2515
    "BOX DRAWINGS UP HEAVY AND RIGHT LIGHT": "┖",  # U+2516
    "BOX DRAWINGS HEAVY UP AND RIGHT": "┗",  # U+2517
    "BOX DRAWINGS LIGHT UP AND LEFT": "┘",  # U+2518
    "BOX DRAWINGS UP LIGHT AND LEFT HEAVY": "┙",  # U+2519
    "BOX DRAWINGS UP HEAVY AND LEFT LIGHT": "┚",  # U+251A
    "BOX DRAWINGS HEAVY UP AND LEFT": "┛",  # U+251B
    "BOX DRAWINGS LIGHT VERTICAL AND RIGHT": "├",  # U+251C
    "BOX DRAWINGS VERTICAL LIGHT AND RIGHT HEAVY": "┝",  # U+251D
    "BOX DRAWINGS UP HEAVY AND RIGHT DOWN LIGHT": "┞",  # U+251E
    "BOX DRAWINGS DOWN HEAVY AND RIGHT UP LIGHT": "┟",  # U+251F
    "BOX DRAWINGS VERTICAL HEAVY AND RIGHT LIGHT": "┠",  # U+2520
    "BOX DRAWINGS DOWN LIGHT AND RIGHT UP HEAVY": "┡",  # U+2521
    "BOX DRAWINGS UP LIGHT AND RIGHT DOWN HEAVY": "┢",  # U+2522
    "BOX DRAWINGS HEAVY VERTICAL AND RIGHT": "┣",  # U+2523
    "BOX DRAWINGS LIGHT VERTICAL AND LEFT": "┤",  # U+2524
    "BOX DRAWINGS VERTICAL LIGHT AND LEFT HEAVY": "┥",  # U+2525
    "BOX DRAWINGS UP HEAVY AND LEFT DOWN LIGHT": "┦",  # U+2526
    "BOX DRAWINGS DOWN HEAVY AND LEFT UP LIGHT": "┧",  # U+2527
    "BOX DRAWINGS VERTICAL HEAVY AND LEFT LIGHT": "┨",  # U+2528
    "BOX DRAWINGS DOWN LIGHT AND LEFT UP HEAVY": "┩",  # U+2529
    "BOX DRAWINGS UP LIGHT AND LEFT DOWN HEAVY": "┪",  # U+252A
    "BOX DRAWINGS HEAVY VERTICAL AND LEFT": "┫",  # U+252B
    "BOX DRAWINGS LIGHT DOWN AND HORIZONTAL": "┬",  # U+252C
    "BOX DRAWINGS LEFT HEAVY AND RIGHT DOWN LIGHT": "┭",  # U+252D
    "BOX DRAWINGS RIGHT HEAVY AND LEFT DOWN LIGHT": "┮",  # U+252E
    "BOX DRAWINGS DOWN LIGHT AND HORIZONTAL HEAVY": "┯",  # U+252F
    "BOX DRAWINGS DOWN HEAVY AND HORIZONTAL LIGHT": "┰",  # U+2530
    "BOX DRAWINGS RIGHT LIGHT AND LEFT DOWN HEAVY": "┱",  # U+2531
    "BOX DRAWINGS LEFT LIGHT AND RIGHT DOWN HEAVY": "┲",  # U+2532
    "BOX DRAWINGS HEAVY DOWN AND HORIZONTAL": "┳",  # U+2533
    "BOX DRAWINGS LIGHT UP AND HORIZONTAL": "┴",  # U+2534
    "BOX DRAWINGS LEFT HEAVY AND RIGHT UP LIGHT": "┵",  # U+2535
    "BOX DRAWINGS RIGHT HEAVY AND LEFT UP LIGHT": "┶",  # U+2536
    "BOX DRAWINGS UP LIGHT AND HORIZONTAL HEAVY": "┷",  # U+2537
    "BOX DRAWINGS UP HEAVY AND HORIZONTAL LIGHT": "┸",  # U+2538
    "BOX DRAWINGS RIGHT LIGHT AND LEFT UP HEAVY": "┹",  # U+2539
    "BOX DRAWINGS LEFT LIGHT AND RIGHT UP HEAVY": "┺",  # U+253A
    "BOX DRAWINGS HEAVY UP AND HORIZONTAL": "┻",  # U+253B
    "BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL": "┼",  # U+253C
    "BOX DRAWINGS LEFT HEAVY AND RIGHT VERTICAL LIGHT": "┽",  # U+253D
    "BOX DRAWINGS RIGHT HEAVY AND LEFT VERTICAL LIGHT": "┾",  # U+253E
    "BOX DRAWINGS VERTICAL LIGHT AND HORIZONTAL HEAVY": "┿",  # U+253F
    "BOX DRAWINGS UP HEAVY AND DOWN HORIZONTAL LIGHT": "╀",  # U+2540
    "BOX DRAWINGS DOWN HEAVY AND UP HORIZONTAL LIGHT": "╁",  # U+2541
    "BOX DRAWINGS VERTICAL HEAVY AND HORIZONTAL LIGHT": "╂",  # U+2542
    "BOX DRAWINGS LEFT UP HEAVY AND RIGHT DOWN LIGHT": "╃",  # U+2543
    "BOX DRAWINGS RIGHT UP HEAVY AND LEFT DOWN LIGHT": "╄",  # U+2544
    "BOX DRAWINGS LEFT DOWN HEAVY AND RIGHT UP LIGHT": "╅",  # U+2545
    "BOX DRAWINGS RIGHT DOWN HEAVY AND LEFT UP LIGHT": "╆",  # U+2546
    "BOX DRAWINGS DOWN LIGHT AND UP HORIZONTAL HEAVY": "╇",  # U+2547
    "BOX DRAWINGS UP LIGHT AND DOWN HORIZONTAL HEAVY": "╈",  # U+2548
    "BOX DRAWINGS RIGHT LIGHT AND LEFT VERTICAL HEAVY": "╉",  # U+2549
    "BOX DRAWINGS LEFT LIGHT AND RIGHT VERTICAL HEAVY": "╊",  # U+254A
    "BOX DRAWINGS HEAVY VERTICAL AND HORIZONTAL": "╋",  # U+254B
    "BOX DRAWINGS LIGHT DOUBLE DASH HORIZONTAL": "╌",  # U+254C
    "BOX DRAWINGS HEAVY DOUBLE DASH HORIZONTAL": "╍",  # U+254D
    "BOX DRAWINGS LIGHT DOUBLE DASH VERTICAL": "╎",  # U+254E
    "BOX DRAWINGS HEAVY DOUBLE DASH VERTICAL": "╏",  # U+254F
    "BOX DRAWINGS DOUBLE HORIZONTAL": "═",  # U+2550
    "BOX DRAWINGS DOUBLE VERTICAL": "║",  # U+2551
    "BOX DRAWINGS DOWN SINGLE AND RIGHT DOUBLE": "╒",  # U+2552
    "BOX DRAWINGS DOWN DOUBLE AND RIGHT SINGLE": "╓",  # U+2553
    "BOX DRAWINGS DOUBLE DOWN AND RIGHT": "╔",  # U+2554
    "BOX DRAWINGS DOWN SINGLE AND LEFT DOUBLE": "╕",  # U+2555
    "BOX DRAWINGS DOWN DOUBLE AND LEFT SINGLE": "╖",  # U+2556
    "BOX DRAWINGS DOUBLE DOWN AND LEFT": "╗",  # U+2557
    "BOX DRAWINGS UP SINGLE AND RIGHT DOUBLE": "╘",  # U+2558
    "BOX DRAWINGS UP DOUBLE AND RIGHT SINGLE": "╙",  # U+2559
    "BOX DRAWINGS DOUBLE UP AND RIGHT": "╚",  # U+255A
    "BOX DRAWINGS UP SINGLE AND LEFT DOUBLE": "╛",  # U+255B
    "BOX DRAWINGS UP DOUBLE AND LEFT SINGLE": "╜",  # U+255C
    "BOX DRAWINGS DOUBLE UP AND LEFT": "╝",  # U+255D
    "BOX DRAWINGS VERTICAL SINGLE AND RIGHT DOUBLE": "╞",  # U+255E
    "BOX DRAWINGS VERTICAL DOUBLE AND RIGHT SINGLE": "╟",  # U+255F
    "BOX DRAWINGS DOUBLE VERTICAL AND RIGHT": "╠",  # U+2560
    "BOX DRAWINGS VERTICAL SINGLE AND LEFT DOUBLE": "╡",  # U+2561
    "BOX DRAWINGS VERTICAL DOUBLE AND LEFT SINGLE": "╢",  # U+2562
    "BOX DRAWINGS DOUBLE VERTICAL AND LEFT": "╣",  # U+2563
    "BOX DRAWINGS DOWN SINGLE AND HORIZONTAL DOUBLE": "╤",  # U+2564
    "BOX DRAWINGS DOWN DOUBLE AND HORIZONTAL SINGLE": "╥",  # U+2565
    "BOX DRAWINGS DOUBLE DOWN AND HORIZONTAL": "╦",  # U+2566
    "BOX DRAWINGS UP SINGLE AND HORIZONTAL DOUBLE": "╧",  # U+2567
    "BOX DRAWINGS UP DOUBLE AND HORIZONTAL SINGLE": "╨",  # U+2568
    "BOX DRAWINGS DOUBLE UP AND HORIZONTAL": "╩",  # U+2569
    "BOX DRAWINGS VERTICAL SINGLE AND HORIZONTAL DOUBLE": "╪",  # U+256A
    "BOX DRAWINGS VERTICAL DOUBLE AND HORIZONTAL SINGLE": "╫",  # U+256B
    "BOX DRAWINGS DOUBLE VERTICAL AND HORIZONTAL": "╬",  # U+256C
    "BOX DRAWINGS LIGHT ARC DOWN AND RIGHT": "╭",  # U+256D
    "BOX DRAWINGS LIGHT ARC DOWN AND LEFT": "╮",  # U+256E
    "BOX DRAWINGS LIGHT ARC UP AND LEFT": "╯",  # U+256F
    "BOX DRAWINGS LIGHT ARC UP AND RIGHT": "╰",  # U+2570
    "BOX DRAWINGS LIGHT DIAGONAL UPPER RIGHT TO LOWER LEFT": "╱",  # U+2571
    "BOX DRAWINGS LIGHT DIAGONAL UPPER LEFT TO LOWER RIGHT": "╲",  # U+2572
    "BOX DRAWINGS LIGHT DIAGONAL CROSS": "╳",  # U+2573
    "BOX DRAWINGS LIGHT LEFT": "╴",  # U+2574
    "BOX DRAWINGS LIGHT UP": "╵",  # U+2575
    "BOX DRAWINGS LIGHT RIGHT": "╶",  # U+2576
    "BOX DRAWINGS LIGHT DOWN": "╷",  # U+2577
    "BOX DRAWINGS HEAVY LEFT": "╸",  # U+2578
    "BOX DRAWINGS HEAVY UP": "╹",  # U+2579
    "BOX DRAWINGS HEAVY RIGHT": "╺",  # U+257A
    "BOX DRAWINGS HEAVY DOWN": "╻",  # U+257B
    "BOX DRAWINGS LIGHT LEFT AND HEAVY RIGHT": "╼",  # U+257C
    "BOX DRAWINGS LIGHT UP AND HEAVY DOWN": "╽",  # U+257D
    "BOX DRAWINGS HEAVY LEFT AND LIGHT RIGHT": "╾",  # U+257E
    "BOX DRAWINGS HEAVY UP AND LIGHT DOWN": "╿",  # U+257F
    # Unicode Block Elements, 0x2580 - 0x259F
    "FULL BLOCK": "█",  # U+2588
    "LEFT SEVEN EIGHTS BLOCK": "▉",  # U+2589
    "LEFT THREE QUARTERS BLOCK": "▊",  # U+258A
    "LEFT FIVE EIGHTS BLOCK": "▋",  # U+258B
    "LEFT HALF BLOCK": "▌",  # U+258C
    "LEFT THREE EIGHTS BLOCK": "▍",  # U+258D
    "LEFT ONE QUARTER BLOCK": "▎",  # U+258E
    "LEFT ONE EIGHT BLOCK": "▏",  # U+258F
    "LIGHT SHADE": "░",  # U+2591
    "MEDIUM SHADE": "▒",  # U+2592
    "DARK SHADE": "▓",  # U+2593
}  # UNICODE_CHARACTERS

# A namespace containing namespaces mapped to Unicode blocks.
# The Unicode blocks contain Unicode symbols by name, the value
# is the symbol itself.
UC = Namespace(
    c0_controls=Namespace(
        nul=unicodedata.lookup('NUL'),
    ), # c0_controls
    general_punctuation=Namespace(
        em_space=unicodedata.lookup('EM SPACE'),
    ),  # general_punctuation
    arrows=Namespace(
        leftwards_arrow=unicodedata.lookup('LEFTWARDS ARROW'),
        upwards_arrow=unicodedata.lookup('UPWARDS ARROW'),
        rightwards_arrow=unicodedata.lookup('RIGHTWARDS ARROW'),
        downwards_arrow=unicodedata.lookup('DOWNWARDS ARROW'),
    ),  # arrows
    box_drawings=Namespace(
        double=Namespace(
            down_and_horizontal=unicodedata.lookup('BOX DRAWINGS DOUBLE DOWN AND HORIZONTAL'),
            down_and_left=unicodedata.lookup('BOX DRAWINGS DOUBLE DOWN AND LEFT'),
            down_and_right=unicodedata.lookup('BOX DRAWINGS DOUBLE DOWN AND RIGHT'),
            horizontal=unicodedata.lookup('BOX DRAWINGS DOUBLE HORIZONTAL'),
            up_and_horizontal=unicodedata.lookup('BOX DRAWINGS DOUBLE UP AND HORIZONTAL'),
            up_and_left=unicodedata.lookup('BOX DRAWINGS DOUBLE UP AND LEFT'),
            up_and_right=unicodedata.lookup('BOX DRAWINGS DOUBLE UP AND RIGHT'),
            vertical=unicodedata.lookup('BOX DRAWINGS DOUBLE VERTICAL'),
            vertical_and_horizontal=unicodedata.lookup('BOX DRAWINGS DOUBLE VERTICAL AND HORIZONTAL'),
            vertical_and_left=unicodedata.lookup('BOX DRAWINGS DOUBLE VERTICAL AND LEFT'),
            vertical_and_right=unicodedata.lookup('BOX DRAWINGS DOUBLE VERTICAL AND RIGHT'),
        ),  # double
        light=Namespace(
            down_and_horizontal=unicodedata.lookup('BOX DRAWINGS LIGHT DOWN AND HORIZONTAL'),
            down_and_left=unicodedata.lookup('BOX DRAWINGS LIGHT DOWN AND LEFT'),
            down_and_right=unicodedata.lookup('BOX DRAWINGS LIGHT DOWN AND RIGHT'),
            horizontal=unicodedata.lookup('BOX DRAWINGS LIGHT HORIZONTAL'),
            up_and_horizontal=unicodedata.lookup('BOX DRAWINGS LIGHT UP AND HORIZONTAL'),
            up_and_left=unicodedata.lookup('BOX DRAWINGS LIGHT UP AND LEFT'),
            up_and_right=unicodedata.lookup('BOX DRAWINGS LIGHT UP AND RIGHT'),
            vertical=unicodedata.lookup('BOX DRAWINGS LIGHT VERTICAL'),
            vertical_and_horizontal=unicodedata.lookup('BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL'),
            vertical_and_left=unicodedata.lookup('BOX DRAWINGS LIGHT VERTICAL AND LEFT'),
            vertical_and_right=unicodedata.lookup('BOX DRAWINGS LIGHT VERTICAL AND RIGHT'),
        ),  # light
        heavy=Namespace(
            down_and_horizontal=unicodedata.lookup('BOX DRAWINGS HEAVY DOWN AND HORIZONTAL'),
            down_and_left=unicodedata.lookup('BOX DRAWINGS HEAVY DOWN AND LEFT'),
            down_and_right=unicodedata.lookup('BOX DRAWINGS HEAVY DOWN AND RIGHT'),
            horizontal=unicodedata.lookup('BOX DRAWINGS HEAVY HORIZONTAL'),
            up_and_horizontal=unicodedata.lookup('BOX DRAWINGS HEAVY UP AND HORIZONTAL'),
            up_and_left=unicodedata.lookup('BOX DRAWINGS HEAVY UP AND LEFT'),
            up_and_right=unicodedata.lookup('BOX DRAWINGS HEAVY UP AND RIGHT'),
            vertical=unicodedata.lookup('BOX DRAWINGS HEAVY VERTICAL'),
            vertical_and_horizontal=unicodedata.lookup('BOX DRAWINGS HEAVY VERTICAL AND HORIZONTAL'),
            vertical_and_left=unicodedata.lookup('BOX DRAWINGS HEAVY VERTICAL AND LEFT'),
            vertical_and_right=unicodedata.lookup('BOX DRAWINGS HEAVY VERTICAL AND RIGHT'),
        ),  # heavy
    ),  # box_drawings
    block_elements=Namespace(
        full_block=unicodedata.lookup('FULL BLOCK'),
        left_seven_eights_block=unicodedata.lookup('LEFT SEVEN EIGHTHS BLOCK'),
        left_three_quarters_block=unicodedata.lookup('LEFT THREE QUARTERS BLOCK'),
        left_five_eights_block=unicodedata.lookup('LEFT FIVE EIGHTHS BLOCK'),
        left_half_block=unicodedata.lookup('LEFT HALF BLOCK'),
        left_three_eights_block=unicodedata.lookup('LEFT THREE EIGHTHS BLOCK'),
        left_one_quarter_block=unicodedata.lookup('LEFT ONE QUARTER BLOCK'),
        left_one_eight_block=unicodedata.lookup('LEFT ONE EIGHTH BLOCK'),
        light_shade=unicodedata.lookup('LIGHT SHADE'),
        medium_shade=unicodedata.lookup('MEDIUM SHADE'),
        dark_shade=unicodedata.lookup('DARK SHADE'),
    ),  # block_elements
)  # UC



__title__ = "pymisclib"
__description__ = "Collection of helpful stuff."
__url__ = "https://rot.regenbogenkraut.de/cd/pymisclib"
__version__ = "1.5.0"
__author__ = "Christian Dönges"
__author_email__ = "cd.pymisclib@platypus-projects.net"
__license__ = "Apache 2.0"
__copyright__ = "© Copyright 2012-2024 by Christian Dönges <cd.pymisclib@platypus-projects.net>"



#!/usr/bin/env python3
# vim ts=4,fileencoding=utf-8
# SPDXVersion: SPDX-2.3
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Copyright 2012-2022, 2023, 2024 by Christian Dönges
# SPDXID: SPDXRef-utilities-py
"""Collection of various utility functions.
"""

# pylint: disable=invalid-name

import argparse
import codecs
import contextlib
import ctypes
import glob
import gzip
import locale
import logging
import os
import re
import shutil
import sys
import traceback
import xml.etree.ElementTree as ET
from datetime import datetime, timezone
from pathlib import Path
from pprint import pformat
from time import mktime
from typing import Any
from zipfile import ZipFile, ZipInfo

# Globals (use sparingly)
logger = logging.getLogger(__name__)


def dir_path(path: str) -> Path:
    """Convert the string to a path and return it if it is a directory.

        This function is intended to be used as the `type` argument to :func:`Argparser.add_argument`.

        :param str path: String containing path to check.
        :return: Path to a directory.
        :rtype: Path

        :raise argparse.ArgumentTypeError: The string was not a valid path to a directory.
    """
    if not os.path.isdir(path):
        raise argparse.ArgumentTypeError(
            f'"{path}" is not a valid path to a directory')
    return Path(path)


def exit_hard(returncode: int = 0):
    """Terminate the running application.

        :param int returncode: Exit code to return to spawning shell.
    """
    os._exit(returncode)  # pylint: disable=protected-access


def extract_from_zip_preserving_mtime(zip_file: ZipFile,
                                      zip_info: ZipInfo,
                                      out_path: dir_path):
    """Extract file from a ZIP archive preserving the modification time.

        :param ZipFile zip_file: The archive to extract from.
        :param ZipInfo zip_info: The information about the file to extract.
        :param dir_path out_path: Destination directory for the extracted
            file (must exist).

        :see: https://stackoverflow.com/q/9813243
    """
    zip_file.extract(zip_info, out_path)
    # Add (0, 0, -1) to the time to indicate we do not know if DST was active.
    date_time = mktime(zip_info.date_time + (0, 0, -1))
    # Set ctime and mtime of the extracted file.
    os.utime(out_path / zip_info.filename, (date_time, date_time))


def file_path(path: str) -> Path:
    """Convert the string to a path and return it if it is a file.

        :param str path: String containing path to check.
        :return: Path to a file.
        :rtype: Path

        :raise argparse.ArgumentTypeError: The string did not contain a valid
            path to a file.
    """
    if not os.path.isfile(path):
        raise argparse.ArgumentTypeError(
            f'"{path}" is not a valid path to a file')
    return Path(path)


def get_language() -> str:
    """Determine the language the current user has set their OS to."""
    if os.name == 'posix':
        # BSD, Darwin, and Linux make it easy.
        lang = os.environ['LANG'].split('.')[0]
    elif os.name == 'nt':
        windll = ctypes.windll.kernel32
        lang = locale.windows_locale[windll.GetUserDefaultUILanguage()]
    else:
        raise RuntimeError(f'unknown OS: {os.name}')
    return lang


def hexdump(b: bytes,
            bytes_per_line: int = 16,
            start_offset: int = 0,
            offset_digits: int = 8,
            show_ascii: bool = True) -> str:
    """Generator function to create a pretty representation of the given bytes and return a line per call.

        ```
        first_prefix   00000000  64 65 66 67 68 69 6A 6B  @ABCDEFG
        always_prefix  00000000  64 65 66 67 68 69 6A 6B  @ABCDEFG
        ```

        :param bytes b: The bytes to print.
        :param int bytes_per_line: Number of bytes per line.
        :param int start_offset:   Starting offset for the first byte.
        :param int offset_digits:  Number of digits in the offset.
        :param bool show_ascii: Print ASCII characters after the hex dump if True.
    """
    length = len(b)
    offset = 0
    if length <= bytes_per_line:
        show_offset = False
    else:
        show_offset = True
    while length > 0:
        if length < bytes_per_line:
            bytes_per_line = length
        b_slice = b[offset:offset + bytes_per_line]
        hs = b_slice.hex(' ')
        if show_offset:
            s = f'{offset + start_offset:0{offset_digits}x}  {hs}  '
        else:
            s = f'{hs}  '
        if show_ascii:
            for c in b_slice:
                c = chr(c)
                if c.isprintable():
                    s += c
                else:
                    s += '.'
        length -= bytes_per_line
        offset += bytes_per_line
        yield s


def initialize_console():
    """Initialize the console to work with UTF-8 encoded strings.

        On windows, the console is strange and if output is redirected to a
        file, it gets even stranger. This confuses Python and even though
        PEP 528 solves the problem for interactive consoles, this does not
        help for non-interactive (which redirected streams are).

        The solution for now is to reconfigure the codecs for stdout and
        stderr to always use UTF-8 and replace unmappable characters.
    """
    if os.name == 'nt':
        if sys.stdout.encoding != 'utf-8':
            sys.stdout = codecs.getwriter('utf-8')(sys.stdout.buffer, 'replace')
            print('Reconfigured stdout to use utf-8 encoding.')
        if sys.stderr.encoding != 'utf-8':
            sys.stderr = codecs.getwriter('utf-8')(sys.stderr.buffer, 'replace')
            print('Reconfigured stderr to use utf-8 encoding.')


def iso8601_str(d: datetime,
                seconds_digits: int = 0,
                compact: bool = False,
                utc: bool = False) -> str:
    """Convert a datetime to an ISO8601 string.

        An ISO8601 compliant string is not fully defined, implementations
        have some leeway in the formatting. This function will produce various
        strings, depending on the input parameters.

        The timezone information ("±HHMM") is only present for aware datetime
        objects (e.g. datetime with tzinfo).

        The number of fractional seconds digits is given by seconds_digits. If
        zero digits are requested, the decimal point is left out (e.g.
        "2023-11-21T18:37:31"). If more than 6 digits are requested, only 6
        digits will be output because datetime objects are limited to
        microsecond resolution.

        Compact notation does not separate the different fields of the string
        (e.g. "YYYY-MM-DDTHH:MM:SS" becomes "YYYYMMDDTHHMMSS").

        UTC notation will convert the datetime into UTC. Note that if the
        datetime is not aware, Python will assume it is in localtime. Instead
        of the time zone offset, a "Z" will be appended.

        :param datetime d: Date to convert to a string.
        :param int seconds_digits: Digit of precision of the seconds.
        :param bool compact: Use separators between values if False,
            compact representation if True.
        :param bool utc: Output UTC time with 'Z' suffix instead of timezone offset.
        :return: String representation of the datetime.

        **Example**

        Given the datetime 2023-11-21T18:37:31.123456+0100 and that we are
        running in GMT-02, the following will be output:

        ===== ============== ======= ===== ===============================
        aware seconds_digits compact utc   Result
        ===== ============== ======= ===== ===============================
        Yes   0              False   False 2023-11-21T18:37:31+0100
        Yes   1              False   False 2023-11-21T18:37:31.1+0100
        Yes   6              False   False 2023-11-21T18:37:31.123456+0100
        Yes   7              False   False 2023-11-21T18:37:31.123456+0100
        No    6              False   False 2023-11-21T18:37:31.123456
        Yes   6              True    False 20231121T183731.123456+0100
        Yes   6              False   True  2023-11-21T17:37:31.123456Z
        No    6              False   True  2023-11-21T15:37:31.123456Z
        ===== ============== ======= ===== ===============================

        **Note**: non-aware and utc=True result depends on current

        .. versionadded:: 1.4.1
    """
    if utc:
        d = d.astimezone(tz=timezone.utc)
    if compact:
        s = f'{d.strftime("%Y%m%dT%H%M%S")}'
    else:
        s = f'{d.strftime("%Y-%m-%dT%H:%M:%S")}'
    if seconds_digits > 0:
        if seconds_digits > 6:
            seconds_digits = 6  # datetime is only precise to microseconds
        fractional_seconds = int(d.microsecond / 10**(6-seconds_digits))
        s += f'.{fractional_seconds:0{seconds_digits}d}'
    if utc:
        s += 'Z'
    elif d.tzinfo:
        s += f'{d.strftime("%z")}'
    return s


def logging_add_trace():
    """Add a loglevel TRACE.

        This function should be called only once.
    """
    if hasattr(logging_add_trace, "nr_calls"):
        logging_add_trace.nr_calls += 1
        logging.getLogger(__name__).warning(
            'logging_add_trace() called %d times.',
            logging_add_trace.nr_calls)
        return
    logging_add_trace.nr_calls = 1  # it doesn't exist yet, so initialize it

    if hasattr(logging, 'TRACE'):
        logging.getLogger(__name__).debug('TRACE logging already enabled.')

    # Perform function.
    logging.TRACE = 9
    logging.addLevelName(logging.TRACE, 'TRACE')

    def trace(self, message, *args, **kws):
        """Output message with level TRACE."""
        if self.isEnabledFor(logging.TRACE):
            # Logger takes its '*args' as 'args'.
            self._log(logging.TRACE, message, args, **kws)  # pylint: disable=protected-access

    logging.Logger.trace = trace


def logging_initialize(
        loglevel: int = logging.WARNING,
        log_dir_path: Path = Path('.'),
        log_file_name_format: str = '%P.log',
        log_rotation: int = 0,
        log_compression: int = 5,
        loglevel_console: int = logging.WARNING,
        loglevel_file: int = logging.DEBUG) -> logging.Logger:
    """Initialize the logging interface with the command line options
        passed through the object 'args'. An instance of the root logger is
        returned to the caller.

        If a submodule uses logging.getLogger('somename'), the logger will
        be a child of the root logger and inherit the settings made here.

        :param int loglevel: Loglevel of the logger. Log messages not meeting
            the level requirement are not processed at all.
        :param Path log_dir_path: Path of the directory containing log files.
            None will prevent log file creation.
        :param str log_file_name_format: Format string for the log file name.
            None will prevent log file creation.
        :param int log_rotation: How many logfiles of the same name to keep.
            If set to 0, any existing log file is overwritten. If set to >0,
            that many old log file copies (named <name>.1, <name>.2, etc.)
            are retained.
        :param int log_compression: Zlib compression level to use in compressing
            log. Level 0 is no compression, level 9 is the highest possible.
        :param int loglevel_console: Loglevel filter of the console logger.
        :param int loglevel_file: Loglevel filter of the file logger.
        :return: The root logger instance for the application.
        :rtype: logging.logger

        :note: If the log_file_name_format contains a timestamp, log_rotation
            will only work on other log file copies with the exact same timestamp.
    """
    # Configure the root logger instance.
    global logger   # # pylint: disable=global-statement
    logger = logging.getLogger()
    logger.setLevel(loglevel)

    # Configure the console handler.
    lc = locale.getpreferredencoding()
    if lc.lower() != 'utf-8':
        locale.setlocale(locale.LC_CTYPE, 'C')
    initialize_console()
    ch = logging.StreamHandler()
    ch.setLevel(loglevel_console)
    terse_formatter = logging.Formatter('%(levelname)-8s %(message)s')
    ch.setFormatter(terse_formatter)
    logger.addHandler(ch)
    logger.debug('Logging to console initialized.')

    if log_file_name_format is not None and log_dir_path is not None:
        log_file_name = string_from_format(log_file_name_format)
        log_file_path = log_dir_path / log_file_name
        rotate_file(log_file_name, log_dir_path, log_rotation, log_compression)

        fh = logging.FileHandler(log_file_path, encoding='utf-8', mode='w')
        fh.setLevel(loglevel_file)
        verbose_formatter = logging.Formatter('%(asctime)s %(name)-26s %(levelname)-8s %(message)s')
        fh.setFormatter(verbose_formatter)
        logger.addHandler(fh)
        logger.debug('Logging to file "%s" initialized.', log_file_path)

    if lc.lower() != 'utf-8':
        if locale.getpreferredencoding().lower() == 'utf-8':
            logger.debug('Changed encoding from "%s" to "utf-8".', lc)
        elif lc == locale.getpreferredencoding():
            logger.debug('Failed to change encoding from "%s" to "utf-8".', lc)
        else:
            logger.warning('Failed to change encoding from "%s" to "utf-8", got "%s".',
                           lc, locale.getpreferredencoding())

    return logger


def initialize_logging(args: argparse.Namespace):
    """Initialize the logging interface with the command line options
        passed through the object 'args'. An instance of the root logger is
        returned to the caller.

        If a submodule uses `logging.getLogger('somename')`, the logger will
        be a child of the root logger and inherit the settings made here.

        :param argparse.Namespace args: Namespace containing attributes used as parameters.
        :return: The root logger instance for the application.
        :rtype: logging.logger

        Namespace attributes used:
            * `args.debug` bool - Enable or disable debug mode.
            * `args.logging` str - Loglevel (e.g. CRITICAL, ERROR, etc.)
            * `args.verbose` bool - Enable or disable verbose mode.

        .. deprecated:: 1.2.0
           Use :func:`logging_initialize` and :func:`logging_add_trace()` (if
           needed) instead.
    """

    # Define a new log level 'trace'
    logging_add_trace()

    # args.loglevel contains the string value of the command line option
    # --loglevel.
    numeric_log_level = getattr(logging, args.loglevel.upper(), None)
    if not isinstance(numeric_log_level, int):
        raise ValueError(f'Invalid log level: {args.loglevel}')

    # Configure the root logger instance.
    global logger   # # pylint: disable=global-statement
    logger = logging.getLogger()
    if args.debug:
        logger.setLevel(logging.TRACE)
    else:
        logger.setLevel(numeric_log_level)

    # Create file handler only if debug mode is active.
    if args.debug:
        app_name = sys.argv[0]
        fh = logging.FileHandler(app_name + '.log', encoding='utf-8', mode='w')
        fh.setLevel(logging.TRACE)

    lc = locale.getpreferredencoding()
    if lc != 'utf-8':
        locale.setlocale(locale.LC_CTYPE, 'C')

    # Create a console handler.
    initialize_console()
    ch = logging.StreamHandler()
    ch.setLevel(numeric_log_level)

    # Create a formatter and add it to the handlers.
    terse_formatter = logging.Formatter('%(levelname)-8s %(message)s')
    ch.setFormatter(terse_formatter)
    if args.debug:
        verbose_formatter = logging.Formatter(
            '%(asctime)s %(name)-26s %(levelname)-8s %(message)s')
        fh.setFormatter(verbose_formatter)

    # Add the handlers to the logger.
    logger.addHandler(ch)
    if args.debug:
        logger.addHandler(fh)

    if args.verbose:
        logger.info('Logging initialized.')

    if lc != 'utf-8':
        if locale.getpreferredencoding() == 'utf-8':
            logger.debug('Changed encoding from "%s" to "utf-8".', lc)
        elif lc == locale.getpreferredencoding():
            logger.debug('Failed to change encoding from "%s" to "utf-8".', lc)
        else:
            logger.warning('Failed to change encoding from "%s" to "utf-8", got "%s".',
                           lc, locale.getpreferredencoding())

    logger.debug('Parsed args are: %s', args)
    logger.warning('utilities.initialize_logging() is deprecated, use utilities.logging_initialize() instead.')
    return logger


def is_power_of_two(n: int) -> bool:
    """Return True if the given number *n* is a power of two.

        :param int n: number to check
        :return: True if *n* is a power of two, False otherwise.
        :rtype: bool
    """
    return (n != 0) and ((n & (n - 1)) == 0)


def log_hexdump(fn_logger: logging.Logger,
                b: bytes,
                bytes_per_line: int = 16,
                level: int = logging.DEBUG,
                start_offset: int = 0,
                first_prefix: str = '',
                always_prefix: str = '',
                show_ascii: bool = True):
    """Log a pretty representation of the given bytes to the logger.

        ```
        first_prefix   00000000  64 65 66 67 68 69 6A 6B  @ABCDEFG
        always_prefix  00000000  64 65 66 67 68 69 6A 6B  @ABCDEFG
        ```

        :param logging.Logger fn_logger: The logger to log to.
        :param bytes b: The bytes to print.
        :param int bytes_per_line: The number of bytes per line.
        :param int level: Level for logging (e.g. CRITICAL, ERROR, .. DEBUG)
        :param int start_offset: The starting offset for the first byte.
        :param str first_prefix: A string before the offset on the first line.
        :param str always_prefix: A string that will be printed before every
            line (except the first if first_prefix was specified).
        :param bool show_ascii: Print ASCII characters after the hex dump if True.
    """
    if len(first_prefix) == 0:
        first_prefix = always_prefix
    elif len(first_prefix) > len(always_prefix):
        always_prefix += ' ' * (len(first_prefix) - len(always_prefix))
    elif len(always_prefix) > len(first_prefix):
        first_prefix += ' ' * (len(always_prefix) - len(first_prefix))
    first = True
    for line in hexdump(b,
                        bytes_per_line=bytes_per_line,
                        start_offset=start_offset,
                        offset_digits=8,
                        show_ascii=show_ascii):
        if first:
            fn_logger.log(level, '%s  %s', first_prefix, line)
            first = False
        else:
            fn_logger.log(level, '%s  %s', always_prefix, line)


def log_pp(fn_logger: logging.Logger,
           obj: Any,
           level: int = logging.DEBUG,
           indent: int = 0,
           nesting_indent: int = 1,
           max_width: int = 80,
           compact: bool = False,
           sort_dicts: bool = True):
    """Log an object formatted with pprint.pformat to a logger.
    Mutli-line output is spread to multiple log entries.

    :param logging.Logger fn_logger: Logger instance to log to.
    :param Any obj: Object to log.
    :param int level: Level for logging (e.g logging.DEBUG, logging.INFO, ...)
    :param int indent: Number of spaces to indent all entries.
    :param int nesting_indent: Additional indentation for each object nesting
        level.
    :param int max_width: Maximum width of the formatted object if `indent` is
        zero. The maximum log message length is `max_width + indent`.
    :param bool compact: True to use compact representation, False for pretty.
    :param bool sort_dicts: True to sort dictionaries before output.

    .. csv-table:: log_pp() Parameter correspondence to `pprint.PrettyPrinter <https://docs.python.org/3/library/pprint.html#pprint.PrettyPrinter>`_
        :header: "log_pp", "pprint.PrettyPrinter"
        :align: left

        "`nesting_indent`", "`indent`"
        "`max_width`", "`width`"
        "`compact`", "`compact`"
        "`sort_dicts`", "`sort_dicts`"
    """
    if not fn_logger.isEnabledFor(level):
        return
    fmt = ' ' * indent + '%s'
    for line in pformat(object=obj,
                        indent=nesting_indent,
                        width=max_width,
                        compact=compact,
                        sort_dicts=sort_dicts).splitlines():
        fn_logger.log(level, fmt, line)


def log_stacktrace(fn_logger: logging.Logger,
                   level: int = logging.DEBUG):
    """Log the current stack trace inside or outside an exception.

        :param logging.Logger fn_logger: The logger to log to.
        :param int level: Level for logging (e.g. CRITICAL, ERROR, .. DEBUG)
    """
    # Get the current exception.
    ex = sys.exc_info()[0]
    # Remove this method from the call stack.
    stack = traceback.extract_stack()[:-1]
    if ex is not None:
        # Exception is present, so remove the call of this method.
        del stack[-1]
    fn_logger.log(level, 'Traceback (most recent call last):')
    for s in traceback.format_list(stack):
        fn_logger.log(level, s)
    if ex is not None:
        fn_logger.log(level, traceback.format_exc())


def log_xml(fn_logger: logging.Logger,
            xml_elem: ET.Element | ET.ElementTree,
            level: int = logging.DEBUG,
            indent: int = 0):
    """Log the XML as one or more log entries.

    :param logging.Logger fn_logger: The logger to log to.
    :param etree._Element xml_elem: The XML element to log.
    :param int level: Level for logging (e.g logging.DEBUG, logging.INFO, ...)
    :param int indent: Number of spaces to indent all entries.
    """
    if not fn_logger.isEnabledFor(level):
        # Save lots of processing cycles.
        return

    if isinstance(xml_elem, ET.ElementTree):
        xml_elem = xml_elem.getroot()
    lines = ET.tostringlist(xml_elem, encoding='unicode')
    fmt = ' ' * indent + '%s'
    for line in lines:
        fn_logger.log(level, fmt, line)


def resolve_wildcards(filenames: list[str]) -> list[Path]:
    """Resolve unique Paths from a list containing wildcards.

        :param list[str] filenames: A list of filenames that may contain
             wildcards.
        :return: A list of unique Paths.
        :rtype: list[Path]
    """
    # Resolve wildcards.
    names = []
    for n in filenames:
        i = glob.glob(n)
        if i:
            names.extend(i)
    # Remove duplicates
    names = list(set(names))

    # Convert strings to paths.
    paths = [Path(n) for n in names]
    return paths


def rotate_file(file_name: str,
                file_dir: Path,
                rotation: int,
                compress_level: int = 9,
                fn_logger: logging.Logger = logging.getLogger(__name__)):
    """Rotate a (log-)file.

        The parameter rotation specifies the number of copies that will be
        retained.

        When rotating, file_name.rotation-1 is renamed to file_name.rotation
        for all values of rotation down to 1.

        The initial file (with no .rotation suffix) gains a suffix. If
        compress_level > 0, the initial file is also compressed.

        :param str file_name: Name of the log file.
        :param Path file_dir: Path to the directory containing the log file.
        :param int rotation: Number of copies to rotate. 0 to disable.
        :param int compress_level: Zlib compression level with 0 being no
            compression and 9 the highest possible.
        :param logging.Logger fn_logger: Logger used by the function.
    """
    fn_logger.debug('rotate_log(%s, %s, %d)', file_name, file_dir, rotation)
    if rotation == 0:
        fn_logger.debug('rotate_log(): nothing to rotate.')
        return

    max_rotation = rotation
    while rotation >= 0:
        if rotation == 0:
            candidate = file_dir / file_name
            if not candidate.exists():
                fn_logger.debug('rotate_log(): rotation 0 does not exists. Done.')
                return
            candidate_is_compressed = False
        else:
            # Find previously rotated candidate.
            candidate = file_dir / f'{file_name}.{rotation}'
            candidate_is_compressed = False
            if not candidate.exists():
                candidate = file_dir / f'{file_name}.{rotation}.gz'
                candidate_is_compressed = True
                if not candidate.exists():
                    fn_logger.debug('rotate_log(): rotation %d does not exist.', rotation)
                    rotation -= 1
                    continue

        # We have found a match.
        fn_logger.debug('rotate_log(): found candidate %s.', candidate)
        if rotation == 0:
            # All other rotation copies have been renamed, now the last logfile
            # must be compressed (optional) and renamed.
            compress_path = file_dir / file_name
            if compress_level > 0:
                new_name = f'{file_name}.1.gz'
                new_path = file_dir / new_name
                fn_logger.debug('rotate_log(): compress %s -> %s.', compress_path, new_path)
                with open(compress_path, 'rb') as f_in:
                    # Get the modification time of the file we are rotating
                    # so it will be stored correctly in the gzip archive.
                    statinfo = os.stat(compress_path)
                    with open(new_path, 'wb') as f_out:
                        with gzip.GzipFile(file_name,
                                           'wb',
                                           compress_level,
                                           f_out,
                                           mtime=statinfo.st_mtime) as f_zip:
                            shutil.copyfileobj(f_in, f_zip)

                fn_logger.debug('rotate_log(): unlink %s.', compress_path)
                compress_path.unlink()
            else:
                new_path = file_dir / f'{file_name}.{rotation + 1}'
                fn_logger.debug('rotate_log(): rename %s -> %s', candidate, new_path)
                candidate.rename(new_path)
        elif rotation == max_rotation:
            # We have found the maximum rotation file. Delete it to make room.
            fn_logger.debug('rotate_log(): unlinking old %s.', candidate)
            candidate.unlink()
        elif rotation > 0:
            # Rename existing rotation to next higher.
            if candidate_is_compressed:
                new_path = file_dir / f'{file_name}.{rotation + 1}.gz'
            else:
                new_path = file_dir / f'{file_name}.{rotation + 1}'
            fn_logger.debug('rotate_log(): rename %s -> %s', candidate, new_path)
            candidate.rename(new_path)
        rotation -= 1


def round_down(n: int, m: int) -> int:
    """Round the given number *n* down to the nearest multiple of *m*.

        :param int n: number to round
        :param int m: multiple to round to
        :return: n rounded down to a multiple of m.
        :rtype: int
    """
    return n & ~(m - 1)


def round_up(n: int, m: int) -> int:
    """Round the given number *n* up to the nearest multiple of *m*.

        :param int n: number to round
        :param int m: multiple to round to
        :return: n rounded up to a multiple of m.
        :rtype: int
    """
    return (n + m - 1) & ~(m - 1)


@contextlib.contextmanager
def std_open(filename: str = None,
             mode: str = 'r',
             encoding: str = 'utf-8',
             newline: str | None = None):
    """Open either a file or stdin/stdout for use with `with`.

        If the filename is None or '-' then stdin or stdout (depending on the
        mode) are used.
        Otherwise, the file is used. It is closed when the `with` block is done.

        :param str filename: A filename, '-', or `None`.
        :param str mode: The mode to use for :func:`open`.
        :param str encoding: The encoding to pass to :func:`open`. Defaults to 'utf-8'.
        :param str newline: Universal newline mode. Passed to :func:`open`.

        Example
        -------
        .. code-block:: python

            with std_open(fn, 'r') as f:
                c = f.read()

        .. versionchanged:: 1.3.1
           Default mode changed to 'r' to minimize chances of accidental file clobbering.
        .. versionchanged:: 1.4.2
           Add parameter `newline`.
    """
    if filename and filename != '-':
        fh = open(filename, mode, encoding=encoding, newline=newline)
    else:
        if 'r' in mode:  # pylint: disable=else-if-used
            fh = sys.stdin
        else:  # 'w'
            fh = sys.stdout

    try:
        yield fh
    finally:
        if fh is not sys.stdin and fh is not sys.stdout:
            fh.close()


def string_from_format(fmt: str, timestamp: datetime = None) -> str:
    """Given a strftime()-like format, expand into a string.

        Additional formats:
        - %P - name of the application

        :param str fmt: Format string to apply.
        :param datetime timestamp: Timestamp to format or `None` to use the current time.
        :return: Resulting string.
    """
    fmt2 = ''
    if '%P' in fmt:
        start = 0
        pos = fmt.find('%P', start)
        while pos >= 0:
            length = pos - start
            fmt2 += fmt[start:length] + Path(sys.argv[0]).stem
            start = pos + 2
            pos = fmt.find('%P', start)
        fmt2 += fmt[start:]
    else:
        fmt2 = fmt

    if timestamp is None:
        timestamp = datetime.now()
    return timestamp.strftime(fmt2)


def string_to_snake_case(string: str) -> str:
    """Convert a string to snake_case.

    :param str string: String to convert.
    :rtype str:
    :return: Converted snake_case string.

    .. seealso:: https://stackoverflow.com/a/1176023
    """
    # Edge case of an acronym followed by another word (e.g. "HTTPResponse"
    # -> "HTTP_Response") OR the more normal case of an initial lowercase word
    # followed by a capitalized word (e.g. "getResponse" -> "get_Response").
    string = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', string)
    # Replace double __ with single _ (e.g. __HTTP_Response -> _HTTP_Response).
    string = re.sub('__([A-Z])', r'_\1', string)
    # Normal case of two non-acronyms (e.g. "ResponseCode" -> "Response_Code")
    string = re.sub('([a-z0-9])([A-Z])', r'\1_\2', string)
    # Convert everything to lower case.
    return string.lower()


def true_stem(path: Path) -> str:
    """Return the true stem (e.g. the name without suffixes) of the Path.

        :param Path path: A Path object.
        :rtype: str
        :return: The final file or directory name without any suffixes.

        Example
        -------

        .. code-block:: python

            print(true_stem(Path('/var/tmp/abc.tar.gz')))
            abc

    """
    ts = path.stem
    while path.suffixes:
        ts = path.stem
        path = Path(ts)
    return ts


def wrapping_counter_delta(first: int, second: int, max_value: int) -> int:
    """Calculate the difference between two values of a wrapping counter.

    :param int first: The first counter value.
    :param int second: The second (later) counter value.
    :param int max_value: The maximum value the counter can reach plus one.
        For fixed-length integers, this is 2**<num_bits>.
    :returns: The difference between the two counter values.

    Example for a 4 bit counter (range 0..15)::

         0             0            0             0 1             1
         0             f            0             f 0             f
        |---------------|          |---------------|---------------|
           ^        ^       ===>               ^      ^
           2nd      1st                        1st    2nd

        first  = 0x0c (= 12)
        second = 0x03 (=  3)
        delta = 0x03 + 0x10 - 0x0c = 0x07
                   3 +   16 -   12 =    7

    .. versionadded:: 1.5.0
    """
    if first <= second:
        delta = second - first
    else:
        delta = second + max_value - first
    return delta


if __name__ == '__main__':
    if sys.version_info < (3, 9):
        sys.stderr.write('FATAL ERROR: Python 3.9 or later is required.\n')
        sys.exit(1)



#!/usr/bin/env python3
# vim: fileencoding=utf8
# SPDXVersion: SPDX-2.3
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Copyright 2024 by Christian Dönges
# SPDXID: SPDXRef-ringbuffer-py
"""Implementation of a fixed-size ringbuffer.

.. versionadded: 1.5.0

"""

import logging
import sys
from typing import Any

from pymisclib.utilities import wrapping_counter_delta


class BufferEmptyError(BufferError):
    """Raised when a buffer is empty and an attempt is made to read from it."""
    pass


class BufferFullError(BufferError):
    """Raised when a buffer is full and an attempt is made to write to it."""
    pass


class RingBuffer:
    """A fixed-size ringbuffer.

    New entries are added to the end of the buffer, the oldest entry is read
    first. Once the buffer has reached full capacity, the newest entry to be
    added will overwrite the oldest.

    The capacity should be chosen carefully to ensure that adding new
    entries never overtakes reading old ones (this will raise an exception).
    Simply opting for the largest possible buffer is not usually an optimal
    strategy because the entries are never freed (i.e. the buffer never shrinks
    in size), potentially consuming large amounts of memory.
    """
    def __init__(self, capacity: int,
                 logger: logging.Logger = logging.getLogger(__name__)):
        """
        :param int capacity: Maximum number of entries the buffer can hold.
        :param logging.Logger logger: Logger instance for diagnostics.
        """
        self._capacity = capacity
        self._logger = logger
        self._buffer = []
        self._last_read_offset = -1   # last read was from this offset
        self._last_write_offset = -1  # last write to this offset

    @property
    def capacity(self) -> int:
        """Capacity of the buffer."""
        return self._capacity

    @property
    def num_unread(self) -> int:
        """Number of entries that may be read from the buffer."""
        if self._last_write_offset < 0:
            # No writes have been performed, so there is no content.
            return 0
        if self._last_read_offset < 0:
            # No reads have been performed, so everything is content.
            return self._last_write_offset + 1
        # Both offsets are now [0..self._capacity[.
        delta = wrapping_counter_delta(self._last_read_offset,
                                       self._last_write_offset,
                                       self._capacity)
        return delta

    @property
    def size(self) -> int:
        """Current number of entries in the buffer."""
        return len(self._buffer)

    def add(self, entry: Any):
        """Add an entry to the buffer.

            :raises BufferFullError: if the buffer is full of unread entries.
        """
        if len(self._buffer) < self._capacity:
            self._buffer.append(entry)
            self._last_write_offset += 1
            return

        if self.num_unread >= self._capacity - 1:
            raise BufferFullError('read required')
        write_offset = (self._last_write_offset + 1) % self._capacity
        self._buffer[write_offset] = entry
        self._last_write_offset = write_offset

    def get(self) -> Any:
        """Read the oldest unread entry from the buffer.

            :raises BufferEmptyError: if the buffer contains no unread entries.
        """
        if self.num_unread == 0:
            raise BufferEmptyError('buffer is empty')
        self._last_read_offset = (self._last_read_offset + 1) % self._capacity
        value = self._buffer[self._last_read_offset]
        return value

    def get_unread(self) -> list[Any]:
        """Read all unread entries from the buffer.

            :returns list[Any]: A list of all unread entries. Empty list if
                the buffer is empty.
        """
        nu = self.num_unread
        if nu == 0:
            return []
        if nu == self._capacity:
            return self._buffer
        first_unread = (self._last_read_offset + 1) % self._capacity
        last_unread = (first_unread + nu) % self._capacity
        if last_unread >= first_unread:
            return self._buffer[first_unread:last_unread]
        return self._buffer[first_unread:] + self._buffer[:last_unread]


def main(argv: list) -> int:
    """Do something useful."""
    ring_buffer = RingBuffer(16)
    for i in range(16):
        ring_buffer.add(i)
    return 0


if __name__ == '__main__':
    if sys.version_info < (3, 9):
        print('FATAL ERROR: Python 3.9.x or later is required.')
        sys.exit(1)
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# vim: fileencoding=utf8
# SPDXVersion: SPDX-2.3
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Copyright 2022, 2023 by Christian Dönges
# SPDXID: SPDXRef-littlejonny-py
"""Format output to a text table.

   In case you have been wondering, the name is a reference to
   https://xkcd.com/327/.
"""

# Postpone type hint evaluation until after the entire class has been parsed.
# This was to become standard in 3.10 but will probably never be.
# @see https://stackoverflow.com/a/33533514
# @see https://mail.python.org/archives/list/python-dev@python.org/thread/CLVXXPQ2T2LQ5MP2Y53VVQFCXYWQJHKZ/
from __future__ import annotations


import enum
from dataclasses import dataclass, field
from typing import Any

from pymisclib.ansiterminal import AnsiControl, BgColor, FgColor, TextStyle
from pymisclib.unicodechars import UC


# Box drawing characters.
# The beauty of Namespace makes it possible to access a character like so:
# box_drawings.light.horizontal.
box_drawings = UC.box_drawings

# Type definition: a Text is a collection of lines, each of which is a string.
Text = list[str]
"""Type definition: a Text is a collection of lines, each of which is a string.

    The lines are organized by row, so the first string in the list is the first text line.
"""


@enum.unique
class LineStyle(enum.Enum):
    """Rendering style for lines."""
    double = box_drawings.double
    light = box_drawings.light
    heavy = box_drawings.heavy


@enum.unique
class OutputFormat(enum.Enum):
    """Format for rendering the output."""
    ANSI = enum.auto()  # Render on an ANSI terminal.
    TEXT = enum.auto()  # Render as plain text.


@dataclass
class Point:
    """Coordinate in column (x) / row (y) space."""
    column: int = 1
    row: int = 1

    def __str__(self) -> str:
        """Return a string representation of the instance."""
        return f'({self.column};{self.row})'


@dataclass
class Box:
    """A box consisting of a frame with content.

        Constructor arguments:
            :param int origin_column: Top left corner column.
            :param int origin_row: Top left corner row.
            :param int width: Width in columns.
            :param int height: Height in rows.
            :param LineStyle line_style: Border line style.
            :param str border_style: Border color and font style expressed as ANSI format codes.
            :param str text_style: Content text color and font style expressed as ANSI format codes.
    """
    origin_column: int = 1  # starting x-coordinate
    """X-coordinate of the upper left corner of the box."""
    origin_row: int = 1  # starting y coordinate
    """Y-coordinate of the upper left corner of the box."""
    width: int = 40
    """Width of the box in columns."""
    height: int = 20
    """Height of the box in rows."""
    line_style: LineStyle = LineStyle.light
    """Style of the border characters."""
    border_style: str = FgColor.Black.value + BgColor.White.value
    """Rendering style (color and font style) of the border."""
    text_style: str = FgColor.Blue.value + BgColor.BrightYellow.value + TextStyle.Bold.value
    """Text style (color and font) of the content."""
    _text: Text = field(default_factory=list)

    @classmethod
    def make_from_points(cls, top_left: Point,
                         bottom_right: Point,
                         line_style: LineStyle = LineStyle.light) -> Box:
        """Create a new Box instance with the given top left and bottom right
        points.

        :param Point top_left: starting point in the upper left corner.
        :param Point bottom_right: ending point (inclusive) in the lower right
            corner.
        :param LineStyle line_style: Style to use when rendering the box.
        :return: An initialized instance.
        :rtype Box:
        """
        if top_left.column < 1:
            raise ValueError('left column too small')
        if top_left.row < 1:
            raise ValueError('top row too small')
        if bottom_right.column <= top_left.column:
            raise ValueError('right column too small')
        if bottom_right.row <= top_left.row:
            raise ValueError('bottom column too small')
        width = bottom_right.column - top_left.column
        height = bottom_right.row - top_left.row
        return Box(top_left.row, top_left.column, width, height, line_style)

    @property
    def text(self) -> Text:
        """Text contained in the box."""
        return self._text

    @text.setter
    def text(self, text: Text, truncate: bool = True):
        """Set the text contained in the box."""
        self._text = []
        if len(text) > self.height:
            if truncate:
                edited_text = text[:self.height]
            else:
                raise ValueError('too many lines')
        else:
            edited_text = text
        for line in edited_text:
            if len(line) > self.width:
                if truncate:
                    line = line[:self.width]
                else:
                    ValueError('line too wide')
            else:
                line = line + ' ' * (self.width - len(line))
            self._text.append(line)

    def render(self, output_format: OutputFormat) -> Text:
        """Render the box in the given output_format.

        :param OutputFormat output_format: Rendering format.
        :return: A list of lines rendering the instance.
        :rtype Text:
        """
        if output_format == OutputFormat.ANSI:
            return self._render_ansi()
        elif output_format == OutputFormat.TEXT:
            return self._render_text()
        raise ValueError('unknown output output_format')

    def _render_ansi(self) -> Text:
        """Render the instance on an ANSI terminal."""
        lines = []
        if self.origin_column > 1:
            prefix = AnsiControl.move_cursor_forward(self.origin_column - 1)
        else:
            prefix = ''
        box = self.line_style.value
        s = self.border_style + box.down_and_right + box.horizontal * self.width + \
            box.down_and_left + TextStyle.Reset.value
        if self.origin_column > 1 or self.origin_row > 1:
            s = AnsiControl.move_cursor_to(self.origin_column, self.origin_row) + s
        else:
            s = prefix + s
        lines.append(s)
        s = prefix + self.border_style + box.vertical + \
            self.text_style + (' ' * self.width) + TextStyle.Reset.value + \
            self.border_style + box.vertical + TextStyle.Reset.value
        for i in range(self.height):
            if i < len(self._text):
                lines.append(prefix + self.border_style + box.vertical +
                             self.text_style + self._text[i] + TextStyle.Reset.value +
                             self.border_style + box.vertical + TextStyle.Reset.value)
            else:
                lines.append(s)
        s = prefix + self.border_style + box.up_and_right + \
            box.horizontal * self.width + box.up_and_left + TextStyle.Reset.value
        lines.append(s)

        return lines

    def _render_text(self) -> Text:
        """Render the instance on dumb terminal (e.g. as plain text).

        The origin column and row are created using spaces and blank lines.

        :return: A list of lines, each of which is a string that contains no
            formatting characters.
        :rtype list[str]:
        """
        lines = []
        if self.origin_row > 1:
            for i in range(self.origin_row - 1):
                lines.append('')
        prefix = ' ' * (self.origin_column - 1)
        box = self.line_style.value
        s = prefix + box.down_and_right + box.horizontal * self.width + box.down_and_left
        lines.append(s)
        s = prefix + box.vertical + ' ' * self.width + box.vertical
        for i in range(self.height):
            if i < len(self._text):
                lines.append(prefix + box.vertical + self._text[i] + box.vertical)
            else:
                lines.append(s)
        s = prefix + box.up_and_right + box.horizontal * self.width + box.up_and_left
        lines.append(s)

        return lines


@dataclass
class ColumnTable:
    """Table with data arranged by column."""
    _headings: list[str] = field(default_factory=list)  # row or column heading
    _formats: list[str] = field(default_factory=list)  # cell formats
    _cells: list[list[Any]] = field(default_factory=list)  # column[row]
    _num_columns: int = 0
    _num_rows: int = 0

    @property
    def cell_formats(self) -> list[str]:
        """Return cell formats by columns."""
        return self._formats

    @property
    def headings(self) -> list[str]:
        """Return headings by column."""
        return self._headings

    @property
    def num_columns(self):
        """Number of columns in the table."""
        return self._num_columns

    @property
    def num_rows(self):
        """Number of rows in the table."""
        return self._num_rows

    def set_table(self,
                  headings: list[str],
                  cell_formats: list[str],
                  cells: list[list[str]]):
        """Set the content of the table with cells specified by column.

            :param list[str] headings: List of headings.
            :param list[str] cell_formats: List of cell formats for each column.
                The formats use the same format-language as Python format strings.
            :param list[str] cells: Matrix of cell contents, specified column[row].

            Example
            ---------

            .. code-block:: python

                ct2 = ColumnTable()
                ct2.set_table(
                    headings=['First', 'Second', 'Third', 'Fourth'],
                    cell_formats=['3d', '08x', 's', '>s'],
                    cells=[[1, 2, 12, 123, 1000],
                           [0x12345678, 0xffffee01, -1, 0, 23],
                           ['abc def geh', '12345', 'minus one', 'zero', 'First'],
                           ['this is a sample text', 'Short', 'negative hex number',
                            'Zero hexadecimal number', 'The first column is too large.']]
                )
                print_lines(ct2.draw(LineStyle.light))

            Output::

                ┌───────┬──────────┬─────────────┬────────────────────────────────┐
                │ First │ Second   │ Third       │ Fourth                         │
                ├───────┼──────────┼─────────────┼────────────────────────────────┤
                │   1   │ 12345678 │ abc def geh │ this is a sample text          │
                │   2   │ ffffee01 │ 12345       │ Short                          │
                │  12   │ -0000001 │ minus one   │ negative hex number            │
                │ 123   │ 00000000 │ zero        │ Zero hexadecimal number        │
                │ 1000  │ 00000017 │ First       │ The first column is too large. │
                └───────┴──────────┴─────────────┴────────────────────────────────┘
        """
        if len(headings) != len(cell_formats):
            raise ValueError('number columns in headings and formats do not match')
        if len(headings) != len(cells):
            raise ValueError('number of columns in headings and cells do not match')
        self._num_columns = len(headings)
        self._num_rows = len(cells[0])
        self._headings = headings
        self._formats = cell_formats
        self._cells = cells

    def set_table_transposed(self,
                             headings: list[str],
                             cell_formats: list[str],
                             cells: list[list[str]]):
        """Set the content of the table with cells specified by row.

            :param list[str] headings: List of headings.
            :param list[str] cell_formats: List of cell formats for each column.
            :param list[str] cells: Matrix of cell contents, specified row[column].

            Example
            ---------

            .. code-block:: python

                ct1 = ColumnTable()
                ct1.set_table_transposed(
                    headings=['First', 'Second', 'Third', 'Fourth'],
                    cell_formats=['3d', '08x', 's', '>s'],
                    cells=[[1, 0x12345678, 'abc def geh', 'this is a sample text'],
                           [2, 0xffffee01, '12345', 'Short'],
                           [12, -1, 'minus one', 'negative hex number'],
                           [123, 0, 'zero', 'Zero hexadecimal number'],
                           [1000, 23, 'First', 'The first column is too large.']]
                )
                print_lines(ct1.draw(LineStyle.light))

            Output::

                ┌───────┬──────────┬─────────────┬────────────────────────────────┐
                │ First │ Second   │ Third       │ Fourth                         │
                ├───────┼──────────┼─────────────┼────────────────────────────────┤
                │   1   │ 12345678 │ abc def geh │ this is a sample text          │
                │   2   │ ffffee01 │ 12345       │ Short                          │
                │  12   │ -0000001 │ minus one   │ negative hex number            │
                │ 123   │ 00000000 │ zero        │ Zero hexadecimal number        │
                │ 1000  │ 00000017 │ First       │ The first column is too large. │
                └───────┴──────────┴─────────────┴────────────────────────────────┘
        """
        transposed_cells = [list(x) for x in zip(*cells)]
        self.set_table(headings, cell_formats, transposed_cells)

    def draw(self, style: LineStyle) -> Text:
        """Render the table to a list of lines.

            :param LineStyle style: The line style to use for the table.
            :rtype Text:
            :return: A list of rows which make up the rendered table.
        """
        formatted_cells = []
        column_widths = []
        for c in range(self.num_columns):
            column_width = len(self._headings[c])
            column_data = self._cells[c]
            fmt = self._formats[c]
            formatted_column = []
            for cell in column_data:
                formatted_cell = f'{cell:{fmt}}'
                formatted_column.append(formatted_cell)
                column_width = max(column_width,  len(formatted_cell))
            formatted_cells.append(formatted_column)
            column_widths.append(column_width)

        box = style.value
        lines = []
        first = True
        s = ''
        for c in range(self.num_columns):
            column_width = column_widths[c]
            if first:
                s = box.down_and_right + box.horizontal * (column_width + 2)
                first = False
            else:
                s += box.down_and_horizontal + box.horizontal * (column_width + 2)
        s += box.down_and_left
        lines.append(s)

        s = ''
        for c in range(self.num_columns):
            column_width = column_widths[c]
            s += box.vertical + f' {self._headings[c]:{column_width}s} '
        s += box.vertical
        lines.append(s)

        first = True
        s = ''
        for c in range(self.num_columns):
            column_width = column_widths[c]
            if first:
                s = box.vertical_and_right + box.horizontal * (column_width + 2)
                first = False
            else:
                s += box.vertical_and_horizontal + box.horizontal * (column_width + 2)
        s += box.vertical_and_left
        lines.append(s)

        for r in range(self.num_rows):
            s = ''
            for c in range(self.num_columns):
                column_width = column_widths[c]
                s += box.vertical + f' {formatted_cells[c][r]:{column_width}s} '
            s += box.vertical
            lines.append(s)

        first = True
        s = ''
        for c in range(self.num_columns):
            column_width = column_widths[c]
            if first:
                s = box.up_and_right + box.horizontal * (column_width + 2)
                first = False
            else:
                s += box.up_and_horizontal + box.horizontal * (column_width + 2)
        s += box.up_and_left
        lines.append(s)

        return lines


def draw_box(x, y, style) -> Text:
    """Construct a box of size x * y with style."""
    box = style.value

    lines = []
    s = box.down_and_right + box.horizontal * x + box.down_and_left
    lines.append(s)
    s = box.vertical + ' ' * x + box.vertical
    for i in range(y):
        lines.append(s)
    s = box.up_and_right + box.horizontal * x + box.up_and_left
    lines.append(s)

    return lines


def print_lines(lines: Text):
    """Print lines stored as a list of strings to stdout."""
    for line in lines:
        print(line)


if __name__ == '__main__':
    print_lines(draw_box(5, 3, LineStyle.light))
    print_lines(draw_box(1, 1, LineStyle.double))
    print_lines(draw_box(40, 3, LineStyle.heavy))

    b1 = Box(8, 3, 17, 9, LineStyle.light,
             FgColor.Green.value + BgColor.White.value,
             FgColor.Blue.value + BgColor.BrightYellow.value + TextStyle.Italic.value)
    b1.text = [
        '(8;7)      (17;8)',
        '',
        '',
        '',
        '        *       ',
        '',
        '',
        '',
        '(8;16)    (17;16)',
    ]
    b2 = Box(1, 1, 64, 13, LineStyle.double,
             FgColor.Black.value + BgColor.Green.value,
             FgColor.Red.value + BgColor.BrightWhite.value + TextStyle.Bold.value)
    b2.text = [
        '         1         2         3         4         5         6    ',
        '1234567890123456789012345678901234567890123456789012345678901234',
        '',
        '                                                           right',
        'left',
        '                               center',
        UC.box_drawings.light.horizontal * b2.width,
        'ABC DEF GHI JKL MNO PRQ STU VWX YZ!',
        'ABC DEF GHI JKL MNO PRQ STU VWX YZ!',
        UC.box_drawings.light.horizontal * b2.width,
    ]
    print_lines(b1.render(OutputFormat.ANSI))
    print_lines(b2.render(OutputFormat.ANSI))

    print_lines(b1.render(OutputFormat.TEXT))
    print_lines(b2.render(OutputFormat.TEXT))

    # row[column]
    ct1 = ColumnTable()
    ct1.set_table_transposed(
        headings=['First', 'Second', 'Third', 'Fourth'],
        cell_formats=['3d', '08x', 's', '>s'],
        cells=[[1, 0x12345678, 'abc def geh', 'this is a sample text'],
               [2, 0xffffee01, '12345', 'Short'],
               [12, -1, 'minus one', 'negative hex number'],
               [123, 0, 'zero', 'Zero hexadecimal number'],
               [1000, 23, 'First', 'The first column is too large.']]
    )
    print_lines(ct1.draw(LineStyle.light))

    # column[row]
    ct2 = ColumnTable()
    ct2.set_table(
        headings=['First', 'Second', 'Third', 'Fourth'],
        cell_formats=['3d', '08x', 's', '>s'],
        cells=[[1, 2, 12, 123, 1000],
               [0x12345678, 0xffffee01, -1, 0, 23],
               ['abc def geh', '12345', 'minus one', 'zero', 'First'],
               ['this is a sample text', 'Short', 'negative hex number',
                'Zero hexadecimal number', 'The first column is too large.']]
    )
    print_lines(ct2.draw(LineStyle.light))



#!/usr/bin/env python3
# vim: fileencoding=utf8
# SPDXVersion: SPDX-2.3
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Copyright 2023 by Christian Dönges
# SPDXID: SPDXRef-init-py

from .__version__ import (
    __author__,
    __author_email__,
    __copyright__,
    __description__,
    __license__,
    __title__,
    __url__,
    __version__,
)



#!/usr/bin/env python3
# vim: fileencoding=utf8
# SPDXVersion: SPDX-2.3
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Copyright 2023 by Christian Dönges
# SPDXID: SPDXRef-test-utilities-iso8601-str-py

import unittest
from datetime import datetime, timedelta, timezone

from pymisclib.utilities import iso8601_str


class TestUtilitiesIso8601Str(unittest.TestCase):

    def setUp(self):
        self.timezones = [
            timezone(timedelta(hours=12), 'NZST'),
            timezone(timedelta(hours=1), 'CET'),
            timezone(timedelta(hours=0), 'GMT'),
            timezone(timedelta(hours=-8), 'PST'),
            timezone(timedelta(hours=-12), 'GMT+12'),
        ]

    @staticmethod
    def tz_offset_str(tz: timezone) -> str:
        """Convert a timezone offset to a string of the form "±HHMM".

            :param timezone tz: Timezone
            :rtype str:
            :return: Timezone offset as a string.
        """
        tz_offset = tz.utcoffset(None)
        tzo_hours = int(tz_offset.seconds / 3600) + tz_offset.days * 24
        tzo_minutes = int(abs((tz_offset.seconds - int(tz_offset.seconds / 3600) * 3600) / 60))
        tzo_str = f'{tzo_hours:+03d}{tzo_minutes:02d}'
        return tzo_str

    def test_compact_aware(self):
        for tz in self.timezones:
            d = datetime(2023, 11, 21, 8, 5, 31, 123456, tzinfo=tz)
            tzo_str = TestUtilitiesIso8601Str.tz_offset_str(tz)
            for digits in range(0, 7):
                s = iso8601_str(d, seconds_digits=digits, compact=True)
                if digits == 0:
                    expected = f'20231121T080531{tzo_str}'
                elif digits == 6:
                    expected = f'20231121T080531.123456{tzo_str}'
                else:
                    expected = f'20231121T080531.{"123456"[:-6+digits]:s}{tzo_str}'
                self.assertEqual(s, expected)

    def test_compact_naive(self):
        d = datetime(2023, 11, 21, 8, 5, 31, 123456)
        for digits in range(0, 7):
            s = iso8601_str(d, seconds_digits=digits, compact=True)
            if digits == 0:
                expected = '20231121T080531'
            elif digits == 6:
                expected = '20231121T080531.123456'
            else:
                expected = f'20231121T080531.{"123456"[:-6+digits]:s}'
            self.assertEqual(s, expected)

    def test_compact_utc_aware(self):
        for tz in self.timezones:
            d = datetime(2023, 11, 21, 8, 5, 31, 123456, tzinfo=tz)
            d_utc = d.astimezone(timezone.utc)
            for digits in range(0, 7):
                s = iso8601_str(d, seconds_digits=digits, compact=True, utc=True)
                if digits == 0:
                    expected = d_utc.strftime('%Y%m%dT%H%M%SZ')
                elif digits == 6:
                    expected = d_utc.strftime('%Y%m%dT%H%M%S.123456Z')
                else:
                    expected = d_utc.strftime(f'%Y%m%dT%H%M%S.{"123456"[:-6+digits]:s}Z')
                self.assertEqual(s, expected)

    def test_compact_utc_naive(self):
        d = datetime(2023, 11, 21, 8, 5, 31, 123456)
        d_utc = d.astimezone(timezone.utc)
        for digits in range(0, 7):
            s = iso8601_str(d, seconds_digits=digits, compact=True, utc=True)
            if digits == 0:
                expected = d_utc.strftime('%Y%m%dT%H%M%SZ')
            elif digits == 6:
                expected = d_utc.strftime('%Y%m%dT%H%M%S.123456Z')
            else:
                expected = d_utc.strftime(f'%Y%m%dT%H%M%S.{"123456"[:-6+digits]:s}Z')
            self.assertEqual(s, expected)

    def test_full_aware(self):
        for tz in self.timezones:
            d = datetime(2023, 11, 21, 8, 5, 31, 123456, tzinfo=tz)
            tzo_str = TestUtilitiesIso8601Str.tz_offset_str(tz)
            for digits in range(0, 7):
                s = iso8601_str(d, seconds_digits=digits, compact=False)
                if digits == 0:
                    expected = f'2023-11-21T08:05:31{tzo_str}'
                elif digits == 6:
                    expected = f'2023-11-21T08:05:31.123456{tzo_str}'
                else:
                    expected = f'2023-11-21T08:05:31.{"123456"[:-6+digits]:s}{tzo_str}'
                self.assertEqual(s, expected)

    def test_full_naive(self):
        d = datetime(2023, 11, 21, 8, 5, 31, 123456)
        for digits in range(0, 7):
            s = iso8601_str(d, seconds_digits=digits, compact=False)
            if digits == 0:
                expected = '2023-11-21T08:05:31'
            elif digits == 6:
                expected = '2023-11-21T08:05:31.123456'
            else:
                expected = f'2023-11-21T08:05:31.{"123456"[:-6+digits]:s}'
            self.assertEqual(s, expected)

    def test_full_utc_aware(self):
        for tz in self.timezones:
            d = datetime(2023, 11, 21, 8, 5, 31, 123456, tzinfo=tz)
            d_utc = d.astimezone(timezone.utc)
            for digits in range(0, 7):
                s = iso8601_str(d, seconds_digits=digits, compact=False, utc=True)
                if digits == 0:
                    expected = d_utc.strftime('%Y-%m-%dT%H:%M:%SZ')
                elif digits == 6:
                    expected = d_utc.strftime('%Y-%m-%dT%H:%M:%S.123456Z')
                else:
                    expected = d_utc.strftime(f'%Y-%m-%dT%H:%M:%S.{"123456"[:-6+digits]:s}Z')
                self.assertEqual(s, expected)

    def test_full_utc_naive(self):
        d = datetime(2023, 11, 21, 8, 5, 31, 123456)
        d_utc = d.astimezone(timezone.utc)
        for digits in range(0, 7):
            s = iso8601_str(d, seconds_digits=digits, compact=False, utc=True)
            if digits == 0:
                expected = d_utc.strftime('%Y-%m-%dT%H:%M:%SZ')
            elif digits == 6:
                expected = d_utc.strftime('%Y-%m-%dT%H:%M:%S.123456Z')
            else:
                expected = d_utc.strftime(f'%Y-%m-%dT%H:%M:%S.{"123456"[:-6+digits]:s}Z')
            self.assertEqual(s, expected)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# vim ts=4,fileencoding=utf-8
# SPDXVersion: SPDX-2.3
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Copyright 2012-2022, 2023 by Christian Dönges
# SPDXID: SPDXRef-ascii-table-py
"""Output an ASCII table to stdout in various formats.
"""

# pylint: disable=consider-using-assignment-expr

import argparse


ascii_names: dict[int, str] = {
    0: 'NUL',
    1: 'SOH',
    2: 'STX',
    3: 'ETX',
    4: 'EOT',
    5: 'ENQ',
    6: 'ACK',
    7: 'BEL',
    8: 'BS',
    9: 'HT',
    10: 'LF',
    11: 'VT',
    12: 'FF',
    13: 'CR',
    14: 'SO',
    15: 'SI',
    16: 'DC0',
    17: 'DC1',
    18: 'DC2',
    19: 'DC3',
    20: 'DC4',
    21: 'NAK',
    22: 'SYN',
    23: 'ETB',
    24: 'CAN',
    25: 'EM',
    26: 'SUB',
    27: 'ESC',
    28: 'FS',
    29: 'GS',
    30: 'RS',
    31: 'US',
    127: 'DEL',
}
"""Dictionary of integer values and their ASCII names.
   Contains only non-printable ASCII values."""

def text_list():
    """Output ASCII table as Python list."""
    print('ascii_list = [')
    for i in range(256):
        if i in ascii_names:
            c = ascii_names[i]
        else:
            c = str(chr(i))
            if not c.isprintable():
                c = '_'
        if c == "'":
            print(f'    "{c}",  # {i} / 0x{i:02x} / {c}')
        elif c == '\\':
            print(f"    '\\',  # {i} / 0x{i:02x} / BACKSLASH")
        else:
            print(f"    '{c}',  # {i} / 0x{i:02x} / {c}")
    print(']')


def text_table():
    """Output ASCII table as text."""
    h1 = '     |'
    h2 = '=====+'
    for lower_nibble in range(0, 16):
        h1 += f'| {lower_nibble:04b} '
        h2 += '+======'
    body = []
    for upper_nibble in range(0, 16):
        s1 = '     |'
        s2 = f'{upper_nibble:04b} |'
        s3 = '     |'
        s4 = '-----+'
        for lower_nibble in range(0, 16):
            i = (upper_nibble << 4) + lower_nibble
            if i in ascii_names:
                c = ascii_names[i]
            else:
                c = str(chr(i))
                if not c.isprintable():
                    c = '???'
            s1 += f'| {i:#04x} '
            s2 += f'|  {i:3d} '
            s3 += f'| {c:4s} '
            s4 += '+------'
        body.append(s1)
        body.append(s2)
        body.append(s3)
        if upper_nibble < 15:
            body.append(s4)

    print(h1)
    print(h2)
    for line in body:
        print(line)


def html_list():
    """Print an ASCII Python 'list' in HTML format."""
    h1 = '<tr>'
    for lower_nibble in range(0, 16):
        h1 += f'<th>{lower_nibble:04b}</th>'
    h1 += '</tr>'
    body = []
    for upper_nibble in range(0, 16):
        s = f'<tr><th>{upper_nibble:04b}</th>'
        for lower_nibble in range(0, 16):
            i = (upper_nibble << 4) + lower_nibble
            c = ascii_names.get(i, f'&#x{i:02x};')
            s += f'<td>{i:#04x}</br>{i:3d}</br>{c}</td>'
        s += '</tr>'
        body.append(s)

    print('''<!DOCTYPE html>
<html>
<head>
    <title>ASCII Table</title>
</head>
<style>
table {
  border-collapse: collapse;
}
table, th, td {
    border: 1px solid black;
    padding: 8px;
    text-align: center;
}
th {
    border: 2px solid;
    vertical-align: middle;
}
tr {
    vertical-align: top;
}
</style>
<body>
    <h1>ASCII Table</h1>
    <pre>
''')
    print('ascii_list = [')
    for i in range(256):
        if i in ascii_names:
            c = ascii_names[i]
        else:
            c = str(chr(i))
            if not c.isprintable():
                c = f'&#x{i:02x};'
        if c == "'":
            print(f'    "{c}",  # {i} / 0x{i:02x} / {c}\r')
        elif c == '\\':
            print(f"    '\\',  # {i} / 0x{i:02x} / BACKSLASH\r")
        else:
            print(f"    '{c}',  # {i} / 0x{i:02x} / {c}\r")
    print(']')

    print('''
    </pre>
</body>
</html>
''')


def html_table():
    """Print an ASCII table in HTML format."""
    h1 = '<tr>'
    for lower_nibble in range(0, 16):
        h1 += f'<th>{lower_nibble:04b}</th>'
    h1 += '</tr>'
    body = []
    for upper_nibble in range(0, 16):
        s = f'<tr><th>{upper_nibble:04b}</th>'
        for lower_nibble in range(0, 16):
            i = (upper_nibble << 4) + lower_nibble
            c = ascii_names.get(i, f'&#x{i:02x};')
            s += f'<td>{i:#04x}</br>{i:3d}</br>{c}</td>'
        s += '</tr>'
        body.append(s)

    print('''<!DOCTYPE html>
<html>
<head>
    <title>ASCII Table</title>
</head>
<style>
table {
  border-collapse: collapse;
}
table, th, td {
    border: 1px solid black;
    padding: 8px;
    text-align: center;
}
th {
    border: 2px solid;
    vertical-align: middle;
}
tr {
    vertical-align: top;
}
</style>
<body>
    <h1>ASCII Table</h1>
    <table>
    <tr><th rowspan="2">Upper Nibble</th><th colspan="16">Lower Nibble</th></tr>
''')
    print(h1)
    for line in body:
        print(line)
    print('</table>')
    print('''
</body>
</html>
''')


def main():
    """It all happens here."""
    parser = argparse.ArgumentParser(description='Output an ASCII table.')
    parser.add_argument(
        '-f', '--format', default='text', type=str,
        choices=['html', 'list', 'python', 'text'],
        help='Output format for the table.')
    args = parser.parse_args()

    if args.format == 'html':
        html_table()
    elif args.format == 'list':
        html_list()
    elif args.format == 'python':
        text_list()
    else:
        text_table()


# This code is executed if the file is called as a stand-alone command.
if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# vim: fileencoding=utf8
# SPDXVersion: SPDX-2.3
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © Copyright 2023 by Christian Dönges
# SPDXID: SPDXRef-test-utilities-iso8601-str-py

import unittest

from pymisclib import utilities

# Create test cases dynamically.
# See https://stackoverflow.com/a/2799009 for how this works.
test_vectors = [
    ['abc', 'abc'],
    ['ABC', 'abc'],
    ['AbCd', 'ab_cd'],
    ['ABCdef', 'ab_cdef'],
    ['aBcde', 'a_bcde'],
    ['aBCDef', 'a_bc_def'],
    ['HTTPResponse', 'http_response'],
    ['getHTTPHeaders', 'get_http_headers'],
    ['abCdeFgh', 'ab_cde_fgh'],
]  # test_vectors

class TestUtilitiesStringToSnakeCase(unittest.TestCase):
    """Test converting a string to snake_case."""
    pass

def create_tests(pair: list[str, str]):
    def do_test_expected(self):
        self.assertEqual(pair[1], utilities.string_to_snake_case(pair[0]))

    return do_test_expected

for index, pair in enumerate(test_vectors):
    test_method = create_tests(pair)
    test_method.__name__ = (f'test_expected_{index}')
    setattr(TestUtilitiesStringToSnakeCase, test_method.__name__, test_method)


if __name__ == '__main__':
    unittest.main()
